// ESFrameObserverSwarm.java

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;

import swarm.activity.Activity;
import swarm.activity.ActionGroup;
import swarm.activity.ActionGroupImpl;
import swarm.activity.Schedule;
import swarm.activity.ScheduleImpl;

import swarm.objectbase.Swarm;
import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.simtoolsgui.GUISwarm;
import swarm.simtoolsgui.GUISwarmImpl;

import swarm.analysis.EZGraph;
import swarm.analysis.EZGraphImpl;

import swarm.gui.Colormap;
import swarm.gui.ColormapImpl;
import swarm.space.Value2dDisplay;
import swarm.space.Value2dDisplayImpl;
import swarm.gui.ZoomRaster;
import swarm.gui.ZoomRasterImpl;

import swarm.space.Object2dDisplay;
import swarm.space.Object2dDisplayImpl;

import swarm.collections.ListImpl;
import swarm.collections.ArrayImpl;

import java.util.*;


/**
 * The Observer contains the Model and the graphic widgets
 * @author  Pietro Terna and several of his students
 */
public class ESFrameObserverSwarm extends GUISwarmImpl {

    /** the number of layers in our global model; each layer is a model */
    public int layerNumber;

    /** update frequency */
    public int displayFrequency;

    /** the time */
    public int time;

    /** the time at which we stop the simulation (if != 0) */
    public int timeToFinish;

    /** displaying all messages on the console */
    public boolean verboseChoice;

    /** two ActionGroup for sequence of GUI events */
    public ActionGroup displayActions1, displayActions2;

    /** the single Schedule instance */
    public Schedule displaySchedule;

    /** the Swarm we are observing */

    public ArrayImpl eSFrameModelSwarmArray;
    public ESFrameModelSwarm eSFrameModelSwarm;

    /** our graphics */
    public ArrayImpl unitRasterArray; 
    public ArrayImpl unitDisplayArray; 
    public ArrayImpl unitActivityDisplayArray; 
    public ArrayImpl numberOfUnitsGraphArray; 
    public ArrayImpl productionGraphArray;
    public ArrayImpl ratioGraphArray;

    public EZGraphImpl 
	numberOfUnitsGraph=null, 
	productionGraph=null, ratioGraph=null;

    // ratios used in ratioGraph
    public float ratio1, ratio2, ratio3;
 
    // the raster where we see units and the related unitDisplay and zoomFactor */
    public ZoomRaster unitRaster;
    public Object2dDisplay unitDisplay;
    public int zoomFactor;

    // the unitActivityDisplay
    public Value2dDisplay unitActivityDisplay;

    //  repetition in building models
    public int ln;
    // internal position in evaluateRatios()
    public int lnn=0;

    /** the address of the probe, if we need to access it; remember casting
     *  (ESFrameObserverSwarm$1$ESFrameObserverProbeMap)
     */
    Object observerProbe;

    /** Constructor for class */
    public ESFrameObserverSwarm (Zone aZone) {

	super(aZone);

	// fill in the relevant parameters (only two, in this case).
	layerNumber = 1;
	displayFrequency = 1;
	verboseChoice=false;
	zoomFactor=4;
	timeToFinish=0;
	time=0;

	// Build a customized probe map using a `local' subclass
	// (a special kind of Java `inner class') of the
	// EmptyProbeMapImpl class.
	// This building operation produces a separated class file, with
	// name  ESFrameObserverSwarm$1$ESFrameObserverProbeMap.class
        class ESFrameObserverProbeMap extends EmptyProbeMapImpl {
	    public VarProbe probeVariable (String name) {
		return
		    Globals.env.probeLibrary.getProbeForVariable$inClass
		    (name, ESFrameObserverSwarm.this.getClass ());
	    }
	    public MessageProbe probeMessage (String name) {
		return
		    Globals.env.probeLibrary.getProbeForMessage$inClass
		    (name, ESFrameObserverSwarm.this.getClass ());
	    }
	    public void addVar (String name) {
		addProbe (probeVariable (name));
	    }
	    public void addMessage (String name) {
		addProbe (probeMessage (name));
	    }

	    public ESFrameObserverProbeMap (Zone _aZone, Class aClass) {
		super (_aZone, aClass);
		addVar ("layerNumber");
		addVar ("displayFrequency");
		addVar ("verboseChoice");
		addVar ("zoomFactor");
		addVar ("time");
		addVar ("timeToFinish");
	    }
	}
	// Install our custom probeMap class directly into the
	// probeLibrary
	observerProbe = new ESFrameObserverProbeMap (aZone, getClass ());
	Globals.env.probeLibrary.setProbeMap$For
	    ((ESFrameObserverProbeMap) observerProbe, getClass ());
    }

    /** Create the objects used to display the model. */
    public Object buildObjects () {

	Colormap colormap;
	Selector sel;
	super.buildObjects ();
	
	System.out.println("NB, the left and rigth histogram bars are in unit of orders;");
	System.out.println("    the lines in the Production graph are in unit of steps.");

	Globals.env.createArchivedProbeDisplay (this, "eSFrameObserverSwarm");

	// Instruct the control panel to wait for a button event: we
	// halt here until someone hits a control panel button so the
	// user can get a chance to fill in parameters before the
	// simulation runs
	getControlPanel ().setStateStopped ();

	// We create the model that we're actually observing. The
	// model is a subswarm of the observer.

	eSFrameModelSwarmArray = new ArrayImpl (getZone(), layerNumber);


	for (ln=0;ln<layerNumber;ln++)
	    {
		// creating the ESFrameModelSwarm via lispAppArchivers, which
		// refers to the jesopenfoundation.scm file where we can modify parameters
		// in a stable way without compiling again the code
		eSFrameModelSwarm =
		    (ESFrameModelSwarm)Globals.env.lispAppArchiver.getWithZone$key(
						      getZone(), "eSFrameModelSwarm"+ln);
		eSFrameModelSwarm.setLayer(ln);
		eSFrameModelSwarmArray.atOffset$put(ln, eSFrameModelSwarm);

		// instead of using the direct constructor way
		// eSFrameModelSwarm = new ESFrameModelSwarm (getZone ());
		
		// Now create probe objects on the model and ourselves. This gives a
		// simple user interface to let the user change parameters.

		Globals.env.createArchivedProbeDisplay (eSFrameModelSwarm,
							"eSFrameModelSwarm"+ln);
	    }
	

	// Instruct the control panel to wait for a button event: we
	// halt here until someone hits a control panel button so the
	// user can get a chance to fill in parameters before the
	// simulation runs
	getControlPanel ().setStateStopped ();

	// to change the verboseChoice via probe you have to write 'true'
	// or 'false'; not simply 't' or 'f'
	// setting the 'verbose' value in StartESFrame
	StartESFrame.verbose=verboseChoice;

	// First, let the model swarm build its objects.
	for (ln=0; ln<layerNumber;ln++)
	    {
		eSFrameModelSwarm = (ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(ln);
		eSFrameModelSwarm.buildObjects ();
	    }

	// create a colormap, to be used with the raster to display units and their
	// activity
	colormap = new ColormapImpl(getZone());
	colormap.setColor$ToName((byte)0, "yellowgreen");
	colormap.setColor$ToName((byte)1, "black");
	colormap.setColor$ToName((byte)2, "yellow");
	colormap.setColor$ToName((byte)3, "blue");
	colormap.setColor$ToName((byte)4, "red");
	colormap.setColor$ToName((byte)5, "grey");


	// the array of the unit rasters
	unitRasterArray = new ArrayImpl(getZone(), layerNumber);
	// the array of the unit displays
	unitDisplayArray = new ArrayImpl(getZone(), layerNumber);
	// the array of the unit activity displays
	unitActivityDisplayArray = new ArrayImpl(getZone(), layerNumber);

	for (ln=0;ln<layerNumber;ln++)
	    {
		// create a "raster widget", a 2-dimensional display
		// window.  We tell the raster what to do if it dies, give it
		// its colormap, set its zoom factor (its actual size on the
		// display screen), set its virtual dimensions to the size of
		// our world, and give it its title.
		unitRaster = new ZoomRasterImpl (getZone(), "unitRaster"+ln);
		sel = SwarmUtils.getSelector(this, "_unitRasterDeath_");
		unitRaster.enableDestroyNotification$notificationMethod 
		    (this, sel);
		unitRaster.setColormap (colormap);
		unitRaster.setZoomFactor (zoomFactor);
		unitRaster.setWidth$Height(
				   (((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(ln)).
				    getUnitSpace()).
				   getSizeX(),
				   (((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(ln)).
				    getUnitSpace()).
				   getSizeY()
				   );
		unitRaster.setWindowTitle ("Unit Space "+ln);

		// This instructs the raster to digest all the information we
		// have just given it and to initialize itself for display.
		unitRaster.pack();
	
		unitRasterArray.atOffset$put(ln, unitRaster);

		// create an Object2dDisplay that will display the units
		// on the raster. (The Object2dDisplay relies on the objects to send
		// their own draw messages to the raster.)  Think of
		// the unitspace as overlaying the raster.)
		sel = SwarmUtils.getSelector("Unit", "drawSelfOn");
		unitDisplay = new Object2dDisplayImpl(getZone(), unitRaster, 
					     ((ESFrameModelSwarm) eSFrameModelSwarmArray.
					      atOffset(ln)).getUnitSpace(), sel);
		
		// the unit activity display
		unitActivityDisplay = new Value2dDisplayImpl(getZone(), unitRaster, 
							     colormap,
					     ((ESFrameModelSwarm) eSFrameModelSwarmArray.
					      atOffset(ln)).getUnitActivitySpace());


		// tell the Raster to send mouse clicks to the
		// unitDisplay, allowing the user to right-click on the display
		// to probe the individual unit. (The right button is button
		// number 3)
		unitRaster.setButton$Client$Message(3, unitDisplay, 
			   SwarmUtils.getSelector(unitDisplay, "makeProbeAtX$Y"));

		unitDisplayArray.atOffset$put(ln, unitDisplay);
		unitActivityDisplayArray.atOffset$put(ln, unitActivityDisplay);
	    }


	// the array of the graphs of the number of units
	numberOfUnitsGraphArray = new ArrayImpl(getZone(), layerNumber);
       
	for (ln=0; ln<layerNumber; ln++)
	    {
		if(((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(ln))
		   .getCreateNumberOfUnitsGraph())
		    {
			// create the graph of the number of units of each layer of model
			numberOfUnitsGraph = new EZGraphImpl
			    (getZone (), "# of Units in layer "+ln, "Time",
			     "num", "numberOfUnitsGraph"+ln);

			numberOfUnitsGraphArray.atOffset$put(ln, numberOfUnitsGraph);

			Globals.env.setWindowGeometryRecordName
			    (numberOfUnitsGraphArray.atOffset(ln), "numberOfUnitsGraph"+ln);

			// the data for the numberOfUnitsGraph
			numberOfUnitsGraph.createSequence$withFeedFrom$andSelector
			    ("numOfU", (ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(ln),
			     SwarmUtils.getSelector((ESFrameModelSwarm) eSFrameModelSwarmArray.
						    atOffset(ln),
						    "getNumberOfUnits"));
		    }
		else numberOfUnitsGraphArray.atOffset$put(ln, null);
	    }
	
	// the array of the graphs of the production data
	productionGraphArray = new ArrayImpl(getZone(), layerNumber);
	for (ln=0; ln<layerNumber; ln++)
	    {
		if(((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(ln))
		    .getCreateProductionGraph())
		{
		    // create the production graph
		    productionGraph = new EZGraphImpl
			(getZone (), "Production in layer "+ln, "Time",
			 "quantity", "productionGraph"+ln);

		    productionGraphArray.atOffset$put(ln, productionGraph);
		    
		    Globals.env.setWindowGeometryRecordName
			(productionGraph, "productionGraph"+ln);

		    // the data for the productionGraph
		    productionGraph.createSequence$withFeedFrom$andSelector
			("globalProduction", (ESFrameModelSwarm) eSFrameModelSwarmArray.
			 atOffset(ln),
			 SwarmUtils.getSelector((ESFrameModelSwarm) eSFrameModelSwarmArray.
						atOffset(ln),
						"getGlobalProduction"));
		    productionGraph.createSequence$withFeedFrom$andSelector
			("producedFinalGoods", (ESFrameModelSwarm) eSFrameModelSwarmArray.
			 atOffset(ln),
			 SwarmUtils.getSelector((ESFrameModelSwarm) eSFrameModelSwarmArray.
						atOffset(ln),
						"getConcludedOrderSteps"));
		}
		   else productionGraphArray.atOffset$put(ln, null);
	    }

	// the array of the graphs of the ratios
	ratioGraphArray = new ArrayImpl(getZone(), layerNumber);
	for (ln=0; ln<layerNumber; ln++)
	    {
		if(((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(ln))
		    .getCreateRatioGraph())
		    {
			// create the graph of the ratios
			ratioGraph = new EZGraphImpl
			    (getZone (), "Ratios in layer "+ln, "Time",
			     "Quotas", "ratioGraph"+ln);

			ratioGraphArray.atOffset$put(ln, ratioGraph);
			
			Globals.env.setWindowGeometryRecordName
			    (ratioGraph, "ratioGraph"+ln);

			// the data for each ratioGraph (NB ratios are evaluated via a 
			//cycle of calculation in evaluateRatios()
			ratioGraph.createSequence$withFeedFrom$andSelector
			    ("global/potential", this, SwarmUtils.getSelector(this,"getRatio1"));
			ratioGraph.createSequence$withFeedFrom$andSelector
			    ("final/potential", this, SwarmUtils.getSelector(this,"getRatio2"));
			ratioGraph.createSequence$withFeedFrom$andSelector
			    ("final/global", this, SwarmUtils.getSelector(this,"getRatio3"));
		    }
		else ratioGraphArray.atOffset$put(ln, null);
	    }

	return this;
    }
    

  /**
   * Create the actions necessary for the simulation. This is where
   * the schedule is built (but not run!)  Here we create a display
   * schedule - this is used to display the state of the world and
   * check for user input. 
   */

    public Object buildActions () {
	int i;

	super.buildActions();

	// First, let our model swarm build its own schedule.

	for (ln=0;ln<layerNumber;ln++)
	    ((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(ln)).buildActions();

	// Create two ActionGroup for display
	displayActions1 = new ActionGroupImpl (getZone());
	displayActions2 = new ActionGroupImpl (getZone());

	// Add the methods to the ActionGroup

	// Schedule the update of the graphic widgets
	// EZGraph files

	for (ln=0; ln<layerNumber;ln++)
	    if(numberOfUnitsGraphArray.atOffset(ln)!=null)
		displayActions1.createActionTo$message
		    (numberOfUnitsGraphArray.atOffset(ln),
		     SwarmUtils.getSelector((EZGraph) numberOfUnitsGraphArray.atOffset(ln), 
					    "step"));

	for (ln=0; ln<layerNumber;ln++)
	    if(productionGraphArray.atOffset(ln)!=null)
		displayActions1.createActionTo$message
		    (productionGraphArray.atOffset(ln),
		     SwarmUtils.getSelector((EZGraph) productionGraphArray.atOffset(ln), 
					    "step"));

	for (ln=0; ln<layerNumber;ln++)
		{
		    // evaluate ratios
		    displayActions1.createActionTo$message(this,
					    SwarmUtils.getSelector(this,"evaluateRatios"));

		    if(ratioGraphArray.atOffset(ln)!=null)
			displayActions1.createActionTo$message
			    (ratioGraphArray.atOffset(ln),
			     SwarmUtils.getSelector((EZGraph) ratioGraphArray.atOffset(ln), 
						    "step"));
		}

	// Schedule the update of the raster
	// unitActivityDisplay and unitDisplay

	for (ln=0;ln<layerNumber;ln++)
	    {
		displayActions1.createActionTo$message
		    ((Value2dDisplay) unitActivityDisplayArray.atOffset(ln),
		     SwarmUtils.getSelector(unitActivityDisplay, "display"));
		displayActions1.createActionTo$message
		    ((Object2dDisplay) unitDisplayArray.atOffset(ln),
		     SwarmUtils.getSelector(unitDisplay, "display"));
		displayActions1.createActionTo$message
		    ((ZoomRaster) unitRasterArray.atOffset(ln),
		     SwarmUtils.getSelector(unitRaster, "drawSelf"));
	    }

	// Schedule the update of the probe displays
	displayActions2.createActionTo$message
	    (Globals.env.probeDisplayManager,
	     SwarmUtils.getSelector(Globals.env.probeDisplayManager,"update"));

	// Schedule an update for the whole user
	// interface code.  This is crucial: without this, no
	// graphics update and the control panel will be
	// dead. It's best to put it at the end of the display
	// schedule
	displayActions2.createActionTo$message(getActionCache(),
       		      SwarmUtils.getSelector(getActionCache(),"doTkEvents"));

	// set the time
	displayActions2.createActionTo$message(this,
       		      SwarmUtils.getSelector(this,"setTime"));

	// check to finish
	displayActions2.createActionTo$message(this,
       		      SwarmUtils.getSelector(this,"finish"));

	// The display schedule. Note the repeat interval: display
	// is frequently the slowest part of a simulation, so redrawing less
	// frequently can be a help.
	displaySchedule = new ScheduleImpl (getZone (), displayFrequency);

	// insert ActionGroup instance on the repeating Schedule
	// instance

	for (i=0;i<displayFrequency;i++)
	    {
		// displayActions1 is activated only in the last tick of each
		// unit of time (a day, a shift, ...), when some conclusive
		// statistics have to be displayed
		if(i==displayFrequency-1)
		    displaySchedule.at$createAction (i, displayActions1);

		displaySchedule.at$createAction (i, displayActions2);
	    }
	
	return this;
    }

    /**
       activateIn: - activate the schedules so they're ready to run.
       The swarmContext argument has to do with what we were activated
       *in*.  Typically the ObserverSwarm is the top-layer Swarm, so
       it's activated in "null". But other Swarms and Schedules and
       such will be activated inside of us.  */

    public Activity activateIn (Swarm swarmContext) {
	// First, activate ourselves (just pass along the context).
	super.activateIn (swarmContext);

	// Activate the model swarm in ourselves. The model swarm is a
	// subswarm of the observer swarm.
	for (ln=0;ln<layerNumber;ln++)
	    ((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(ln)).activateIn (this);

	// Now activate our schedule in ourselves. This arranges for
	// the execution of the schedule we built.
	displaySchedule.activateIn (this);

	// Activate returns the swarm activity - the thing that's ready to run.
	return getActivity();
  }

    /** drop the Observer, but after having dropped the EZGraph */
    public void drop() {

	System.out.println("\nPress Quit again to exit.");
	getControlPanel ().setStateStopped ();

	// to avoid Tcl/Tk strange errors it is better to drop the graphs before dropping
	// the Observer

	for (ln=0;ln<layerNumber;ln++){
	    numberOfUnitsGraph=(EZGraphImpl) numberOfUnitsGraphArray.atOffset(ln);
	    if(numberOfUnitsGraph!=null)
		numberOfUnitsGraph.drop();
	    numberOfUnitsGraph=null;
	}

	for (ln=0;ln<layerNumber;ln++){
	    productionGraph=(EZGraphImpl) productionGraphArray.atOffset(ln);
	    if(productionGraph!=null)
		productionGraph.drop();
	    productionGraph=null;
	}

	if(ratioGraph!=null)
	    ratioGraph.drop();
	ratioGraph=null;

	for (ln=0;ln<layerNumber;ln++)
	    if(((ZoomRaster) unitRasterArray.atOffset(ln))!=null)
		((ZoomRaster) unitRasterArray.atOffset(ln)).drop ();
	unitRaster=null;

	super.drop(); // here we drop the Observer, dropping its super class
    }

    /** to finish automatically, if timeToFinish!=0 */
    public void finish(){
	if(timeToFinish != 0 && timeToFinish<=Globals.env.getCurrentTime())
	    {
		System.out.println();
		System.out.println("jES virtually quitted at time " + 
		                                   Globals.env.getCurrentTime()+
				   "; press Quit to exit.");
		getControlPanel().setStateStopped();
	    }
    }

    /** handling the raster */
    public Object _unitRasterDeath_ (Object caller) {
	if (unitRaster!=null)
	    unitRaster.drop ();
	unitRaster = null;
	return this;
    }

    /** evaluating ratio1, ratio2, ratio3 for model in layer 'ln' */
    public void evaluateRatios()
    {
	if(lnn/layerNumber == 1) lnn=0;

	// evaluating ratio1 (global / potential production)
	if(((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(lnn)).
	   getPotentialProduction()==0)ratio1=0;
	else ratio1=((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(lnn)).
		 getGlobalProduction() /  
		 ((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(lnn)).
		 getPotentialProduction();

	// evaluating ratio2 (global / potential production)
	if (((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(lnn)).
	    getPotentialProduction()==0)ratio2=0;
	else ratio2=((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(lnn)).
		 getConcludedOrderSteps() /  
		 ((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(lnn)).
		 getPotentialProduction();

	// evaluating ratio2 (final / global production)
	if(((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(lnn)).
	   getGlobalProduction()==0)ratio3=0;
	else ratio3=((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(lnn)).
		 getConcludedOrderSteps() /  
		 ((ESFrameModelSwarm) eSFrameModelSwarmArray.atOffset(lnn)).
		 getGlobalProduction();

	lnn++;
    }

    /** returning ratio1 (global / potential production) */
    public float getRatio1()
    {
	return ratio1;
    }

    /** returning ratio2 (global / potential production) */
    public float getRatio2()
    {
	return ratio2;
    }

    /** returning ratio3 (final / global production) */
    public float getRatio3()
    {
	return ratio3;
    }

    /** setting the time of the simulation */
    public void setTime()
    {
	time=Globals.env.getCurrentTime();
    }

}
